"use strict";
/**
 * Executes PersistOperation in the given connection.
 * @internal
 */
var PersistOperationExecutor = (function () {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    function PersistOperationExecutor(driver, entityMetadatas, broadcaster) {
        this.driver = driver;
        this.entityMetadatas = entityMetadatas;
        this.broadcaster = broadcaster;
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    /**
     * Executes given persist operation.
     */
    PersistOperationExecutor.prototype.executePersistOperation = function (persistOperation) {
        var _this = this;
        return Promise.resolve()
            .then(function () { return _this.broadcastBeforeEvents(persistOperation); })
            .then(function () { return _this.driver.beginTransaction(); })
            .then(function () { return _this.executeInsertOperations(persistOperation); })
            .then(function () { return _this.executeInsertJunctionsOperations(persistOperation); })
            .then(function () { return _this.executeRemoveJunctionsOperations(persistOperation); })
            .then(function () { return _this.executeUpdateRelationsOperations(persistOperation); })
            .then(function () { return _this.executeUpdateOperations(persistOperation); })
            .then(function () { return _this.executeRemoveRelationOperations(persistOperation); })
            .then(function () { return _this.executeRemoveOperations(persistOperation); })
            .then(function () { return _this.driver.endTransaction(); })
            .then(function () { return _this.updateIdsOfInsertedEntities(persistOperation); })
            .then(function () { return _this.updateIdsOfRemovedEntities(persistOperation); })
            .then(function () { return _this.updateSpecialColumnsInEntities(persistOperation); })
            .then(function () { return _this.broadcastAfterEvents(persistOperation); });
    };
    // -------------------------------------------------------------------------
    // Private Methods
    // -------------------------------------------------------------------------
    /**
     * Broadcast all before persistment events - beforeInsert, beforeUpdate and beforeRemove events.
     */
    PersistOperationExecutor.prototype.broadcastBeforeEvents = function (persistOperation) {
        /*console.log("persistOperation.allPersistedEntities: ", persistOperation.allPersistedEntities);
        console.log("inserts", persistOperation.inserts);
        console.log("updates", persistOperation.updates);*/
        var _this = this;
        var insertEvents = persistOperation.inserts.map(function (insertOperation) {
            var persistedEntityWithId = persistOperation.allPersistedEntities.find(function (e) { return e.entity === insertOperation.entity; });
            return _this.broadcaster.broadcastBeforeInsertEvent(persistedEntityWithId.entity);
        });
        var updateEvents = persistOperation.updates.map(function (updateOperation) {
            var persistedEntityWithId = persistOperation.allPersistedEntities.find(function (e) { return e.entity === updateOperation.entity; });
            return _this.broadcaster.broadcastBeforeUpdateEvent(persistedEntityWithId.entity, updateOperation.columns);
        });
        var removeEvents = persistOperation.removes.map(function (removeOperation) {
            // we can send here only dbEntity, not entity from the persisted object, since entity from the persisted
            // object does not exist anymore - its removed, and there is no way to find this removed object
            return _this.broadcaster.broadcastBeforeRemoveEvent(removeOperation.entity, removeOperation.entityId);
        });
        return Promise.all(insertEvents)
            .then(function () { return Promise.all(updateEvents); })
            .then(function () { return Promise.all(removeEvents); }); // todo: do we really should send it in order?
    };
    /**
     * Broadcast all after persistment events - afterInsert, afterUpdate and afterRemove events.
     */
    PersistOperationExecutor.prototype.broadcastAfterEvents = function (persistOperation) {
        var _this = this;
        var insertEvents = persistOperation.inserts.map(function (insertOperation) {
            var persistedEntity = persistOperation.allPersistedEntities.find(function (e) { return e.entity === insertOperation.entity; });
            return _this.broadcaster.broadcastAfterInsertEvent(persistedEntity.entity);
        });
        var updateEvents = persistOperation.updates.map(function (updateOperation) {
            var persistedEntityWithId = persistOperation.allPersistedEntities.find(function (e) { return e.entity === updateOperation.entity; });
            return _this.broadcaster.broadcastAfterUpdateEvent(persistedEntityWithId.entity, updateOperation.columns);
        });
        var removeEvents = persistOperation.removes.map(function (removeOperation) {
            // we can send here only dbEntity, not entity from the persisted object, since entity from the persisted
            // object does not exist anymore - its removed, and there is no way to find this removed object
            return _this.broadcaster.broadcastAfterRemoveEvent(removeOperation.entity, removeOperation.entityId);
        });
        return Promise.all(insertEvents)
            .then(function () { return Promise.all(updateEvents); })
            .then(function () { return Promise.all(removeEvents); }); // todo: do we really should send it in order?
    };
    /**
     * Executes insert operations.
     */
    PersistOperationExecutor.prototype.executeInsertOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.inserts.map(function (operation) {
            return _this.insert(operation).then(function (insertId) {
                operation.entityId = insertId;
            });
        }));
    };
    /**
     * Executes insert junction operations.
     */
    PersistOperationExecutor.prototype.executeInsertJunctionsOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.junctionInserts.map(function (junctionOperation) {
            return _this.insertJunctions(junctionOperation, persistOperation.inserts);
        }));
    };
    /**
     * Executes remove junction operations.
     */
    PersistOperationExecutor.prototype.executeRemoveJunctionsOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.junctionRemoves.map(function (junctionOperation) {
            return _this.removeJunctions(junctionOperation);
        }));
    };
    /**
     * Executes update relations operations.
     */
    PersistOperationExecutor.prototype.executeUpdateRelationsOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.updatesByRelations.map(function (updateByRelation) {
            _this.updateByRelation(updateByRelation, persistOperation.inserts);
        }));
    };
    /**
     * Executes update operations.
     */
    PersistOperationExecutor.prototype.executeUpdateOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.updates.map(function (updateOperation) {
            return _this.update(updateOperation);
        }));
    };
    /**
     * Executes remove relations operations.
     */
    PersistOperationExecutor.prototype.executeRemoveRelationOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.removes
            .filter(function (operation) {
            return !!(operation.relation && !operation.relation.isManyToMany && !operation.relation.isOneToMany);
        })
            .map(function (operation) {
            return _this.updateDeletedRelations(operation);
        }));
    };
    /**
     * Executes remove operations.
     */
    PersistOperationExecutor.prototype.executeRemoveOperations = function (persistOperation) {
        var _this = this;
        return Promise.all(persistOperation.removes.map(function (operation) {
            return _this.delete(operation.entity);
        }));
    };
    /**
     * Updates all ids of the inserted entities.
     */
    PersistOperationExecutor.prototype.updateIdsOfInsertedEntities = function (persistOperation) {
        var _this = this;
        persistOperation.inserts.forEach(function (insertOperation) {
            var metadata = _this.entityMetadatas.findByTarget(insertOperation.entity.constructor);
            insertOperation.entity[metadata.primaryColumn.propertyName] = insertOperation.entityId;
        });
    };
    /**
     * Updates all special columns of the saving entities (create date, update date, versioning).
     */
    PersistOperationExecutor.prototype.updateSpecialColumnsInEntities = function (persistOperation) {
        var _this = this;
        persistOperation.inserts.forEach(function (insertOperation) {
            var metadata = _this.entityMetadatas.findByTarget(insertOperation.entity.constructor);
            if (metadata.updateDateColumn)
                insertOperation.entity[metadata.updateDateColumn.propertyName] = insertOperation.date;
            if (metadata.createDateColumn)
                insertOperation.entity[metadata.createDateColumn.propertyName] = insertOperation.date;
            if (metadata.versionColumn)
                insertOperation.entity[metadata.versionColumn.propertyName]++;
        });
        persistOperation.updates.forEach(function (updateOperation) {
            var metadata = _this.entityMetadatas.findByTarget(updateOperation.entity.constructor);
            if (metadata.updateDateColumn)
                updateOperation.entity[metadata.updateDateColumn.propertyName] = updateOperation.date;
            if (metadata.createDateColumn)
                updateOperation.entity[metadata.createDateColumn.propertyName] = updateOperation.date;
            if (metadata.versionColumn)
                updateOperation.entity[metadata.versionColumn.propertyName]++;
        });
    };
    /**
     * Removes all ids of the removed entities.
     */
    PersistOperationExecutor.prototype.updateIdsOfRemovedEntities = function (persistOperation) {
        var _this = this;
        persistOperation.removes.forEach(function (removeOperation) {
            var metadata = _this.entityMetadatas.findByTarget(removeOperation.entity.constructor);
            var removedEntity = persistOperation.allPersistedEntities.find(function (allNewEntity) {
                return allNewEntity.entity.constructor === removeOperation.entity.constructor && allNewEntity.id === removeOperation.entity[metadata.primaryColumn.name];
            });
            if (removedEntity)
                removedEntity.entity[metadata.primaryColumn.propertyName] = undefined;
        });
    };
    PersistOperationExecutor.prototype.updateByRelation = function (operation, insertOperations) {
        var tableName, relationName, relationId, idColumn, id;
        var relatedInsertOperation = insertOperations.find(function (o) { return o.entity === operation.targetEntity; });
        var idInInserts = relatedInsertOperation ? relatedInsertOperation.entityId : null;
        if (operation.updatedRelation.isOneToMany) {
            var metadata = this.entityMetadatas.findByTarget(operation.insertOperation.entity.constructor);
            tableName = metadata.table.name;
            relationName = operation.updatedRelation.inverseRelation.name;
            relationId = operation.targetEntity[metadata.primaryColumn.propertyName] || idInInserts;
            idColumn = metadata.primaryColumn.name;
            id = operation.insertOperation.entityId;
        }
        else {
            var metadata = this.entityMetadatas.findByTarget(operation.targetEntity.constructor);
            tableName = metadata.table.name;
            relationName = operation.updatedRelation.name;
            relationId = operation.insertOperation.entityId;
            idColumn = metadata.primaryColumn.name;
            id = operation.targetEntity[metadata.primaryColumn.propertyName] || idInInserts;
        }
        return this.driver.update(tableName, (_a = {}, _a[relationName] = relationId, _a), (_b = {}, _b[idColumn] = id, _b));
        var _a, _b;
    };
    PersistOperationExecutor.prototype.update = function (updateOperation) {
        var _this = this;
        var entity = updateOperation.entity;
        var metadata = this.entityMetadatas.findByTarget(entity.constructor);
        var values = updateOperation.columns.reduce(function (object, column) {
            var value = _this.driver.preparePersistentValue(entity[column.propertyName], column);
            object[column.name] = value;
            return object;
        }, {});
        // todo: what if number of updated columns = 0 ? + probably no need to update updated date and version columns
        if (metadata.updateDateColumn)
            values[metadata.updateDateColumn.name] = this.driver.preparePersistentValue(new Date(), metadata.updateDateColumn);
        if (metadata.versionColumn)
            values[metadata.versionColumn.name] = this.driver.preparePersistentValue(entity[metadata.versionColumn.propertyName] + 1, metadata.versionColumn);
        return this.driver.update(metadata.table.name, values, (_a = {}, _a[metadata.primaryColumn.name] = metadata.getEntityId(entity), _a));
        var _a;
    };
    PersistOperationExecutor.prototype.updateDeletedRelations = function (removeOperation) {
        if (removeOperation.relation) {
            return this.driver.update(removeOperation.fromMetadata.table.name, (_a = {}, _a[removeOperation.relation.name] = null, _a), (_b = {}, _b[removeOperation.fromMetadata.primaryColumn.name] = removeOperation.fromEntityId, _b));
        }
        throw new Error("Remove operation relation is not set"); // todo: find out how its possible
        var _a, _b;
    };
    PersistOperationExecutor.prototype.delete = function (entity) {
        var metadata = this.entityMetadatas.findByTarget(entity.constructor);
        return this.driver.delete(metadata.table.name, (_a = {}, _a[metadata.primaryColumn.name] = entity[metadata.primaryColumn.propertyName], _a));
        var _a;
    };
    PersistOperationExecutor.prototype.insert = function (operation) {
        var _this = this;
        var entity = operation.entity;
        var metadata = this.entityMetadatas.findByTarget(entity.constructor);
        var columns = metadata.columns
            .filter(function (column) { return !column.isVirtual; })
            .filter(function (column) { return entity.hasOwnProperty(column.propertyName); })
            .map(function (column) { return column.name; });
        var values = metadata.columns
            .filter(function (column) { return !column.isVirtual; })
            .filter(function (column) { return entity.hasOwnProperty(column.propertyName); })
            .map(function (column) { return _this.driver.preparePersistentValue(entity[column.propertyName], column); });
        var relationColumns = metadata.relations
            .filter(function (relation) { return relation.isOwning && !!relation.inverseEntityMetadata; })
            .filter(function (relation) { return entity.hasOwnProperty(relation.propertyName); })
            .filter(function (relation) { return entity[relation.propertyName][relation.inverseEntityMetadata.primaryColumn.name]; })
            .map(function (relation) { return relation.name; });
        var relationValues = metadata.relations
            .filter(function (relation) { return relation.isOwning && !!relation.inverseEntityMetadata; })
            .filter(function (relation) { return entity.hasOwnProperty(relation.propertyName); })
            .filter(function (relation) { return entity[relation.propertyName].hasOwnProperty(relation.inverseEntityMetadata.primaryColumn.name); })
            .map(function (relation) { return entity[relation.propertyName][relation.inverseEntityMetadata.primaryColumn.name]; });
        var allColumns = columns.concat(relationColumns);
        var allValues = values.concat(relationValues);
        if (metadata.createDateColumn) {
            allColumns.push(metadata.createDateColumn.name);
            allValues.push(this.driver.preparePersistentValue(operation.date, metadata.createDateColumn));
        }
        if (metadata.updateDateColumn) {
            allColumns.push(metadata.updateDateColumn.name);
            allValues.push(this.driver.preparePersistentValue(operation.date, metadata.updateDateColumn));
        }
        if (metadata.versionColumn) {
            allColumns.push(metadata.versionColumn.name);
            allValues.push(this.driver.preparePersistentValue(1, metadata.versionColumn));
        }
        return this.driver.insert(metadata.table.name, this.zipObject(allColumns, allValues));
    };
    PersistOperationExecutor.prototype.insertJunctions = function (junctionOperation, insertOperations) {
        var junctionMetadata = junctionOperation.metadata;
        var metadata1 = this.entityMetadatas.findByTarget(junctionOperation.entity1.constructor);
        var metadata2 = this.entityMetadatas.findByTarget(junctionOperation.entity2.constructor);
        var columns = junctionMetadata.columns.map(function (column) { return column.name; });
        var id1 = junctionOperation.entity1[metadata1.primaryColumn.name] || insertOperations.find(function (o) { return o.entity === junctionOperation.entity1; }).entityId;
        var id2 = junctionOperation.entity2[metadata2.primaryColumn.name] || insertOperations.find(function (o) { return o.entity === junctionOperation.entity2; }).entityId;
        var values = [id1, id2]; // todo: order may differ, find solution (column.table to compare with entity metadata table?)
        return this.driver.insert(junctionMetadata.table.name, this.zipObject(columns, values));
    };
    PersistOperationExecutor.prototype.removeJunctions = function (junctionOperation) {
        var junctionMetadata = junctionOperation.metadata;
        var metadata1 = this.entityMetadatas.findByTarget(junctionOperation.entity1.constructor);
        var metadata2 = this.entityMetadatas.findByTarget(junctionOperation.entity2.constructor);
        var columns = junctionMetadata.columns.map(function (column) { return column.name; });
        var id1 = junctionOperation.entity1[metadata1.primaryColumn.name];
        var id2 = junctionOperation.entity2[metadata2.primaryColumn.name];
        return this.driver.delete(junctionMetadata.table.name, (_a = {}, _a[columns[0]] = id1, _a[columns[1]] = id2, _a));
        var _a;
    };
    PersistOperationExecutor.prototype.zipObject = function (keys, values) {
        return keys.reduce(function (object, column, index) {
            object[column] = values[index];
            return object;
        }, {});
    };
    return PersistOperationExecutor;
}());
exports.PersistOperationExecutor = PersistOperationExecutor;
//# sourceMappingURL=PersistOperationExecutor.js.map